/*
 *  structures.h - System call prototypes
 *  onyx-the-black-cat v0.2
 *
 *  For Mac OS X Tiger 10.4.11 and Leopard 10.5.5
 *
 */

#define PAD_(t) (sizeof(uint64_t) <= sizeof(t) ? \
                0 : sizeof(uint64_t) - sizeof(t))

#if BYTE_ORDER == LITTLE_ENDIAN
#define PADL_(t)        0
#define PADR_(t)        PAD_(t)
#else
#define PADL_(t)        PAD_(t)
#define PADR_(t)        0
#endif

/** ptrace request */
#define PT_DENY_ATTACH 31

/*
Download xnu kernel and find the syscall numbers at 
xnu/bsd/kern/init_sysent.c
or xnu/bsd/sys/syscall.h
*/
/* nosys syscall */
#define SYS_syscall 0
/* exit syscall */
#define SYS_exit 1
/* fork syscall */
#define SYS_fork 2
/* read syscall */
#define SYS_read 3
/* wait4 syscall */
#define SYS_wait4 7
/* ptrace() syscall */
#define SYS_ptrace 26
/* connect() syscall */
#define SYS_connect 98
/* open() syscall */
#define SYS_open 5
/* sysctl() syscall */
#define SYS___sysctl 202
/* chmod() syscall */
#define SYS_chmod 15
/* relevant extended attributes syscalls */
#define SYS_getxattr 234
#define SYS_fgetxattr 235
#define SYS_setxattr 236
#define SYS_fsetxattr 237
#define SYS_removexattr 238
#define SYS_fremovexattr 239
#define SYS_listxattr 240
#define SYS_flistxattr 241

/*
 Find arguments structure at xnu/bsd/sys/sysproto.h
*/
// structure for ptrace() call
struct ptrace_args {
	char req_l_[PADL_(int)]; int req; char req_r_[PADR_(int)];
	char pid_l_[PADL_(pid_t)]; pid_t pid; char pid_r_[PADR_(pid_t)];
	char addr_l_[PADL_(caddr_t)]; caddr_t addr; char addr_r_[PADR_(caddr_t)];
	char data_l_[PADL_(int)]; int data; char data_r_[PADR_(int)];
};

// structure for open() call
struct open_args {
        char path_l_[PADL_(user_addr_t)]; user_addr_t path; char path_r_[PADR_(user_addr_t)];
        char flags_l_[PADL_(int)]; int flags; char flags_r_[PADR_(int)];
        char mode_l_[PADL_(int)]; int mode; char mode_r_[PADR_(int)];
}; 

// structure for sysctl() call (grab this from the xnu, bsd/sys/sysproto.h)
struct sysctl_args {
        char name_l_[PADL_(user_addr_t)]; user_addr_t name; char name_r_[PADR_(user_addr_t)];
        char namelen_l_[PADL_(u_int)]; u_int namelen; char namelen_r_[PADR_(u_int)];
        char old_l_[PADL_(user_addr_t)]; user_addr_t old; char old_r_[PADR_(user_addr_t)];
        char oldlenp_l_[PADL_(user_addr_t)]; user_addr_t oldlenp; char oldlenp_r_[PADR_(user_addr_t)];
        char new_l_[PADL_(user_addr_t)]; user_addr_t new; char new_r_[PADR_(user_addr_t)];
        char newlen_l_[PADL_(user_size_t)]; user_size_t newlen; char newlen_r_[PADR_(user_size_t)];
};
// structure for chmod() call
struct chmod_args {
        char path_l_[PADL_(user_addr_t)]; user_addr_t path; char path_r_[PADR_(user_addr_t)];
        char mode_l_[PADL_(int)]; int mode; char mode_r_[PADR_(int)];
}; 
#if MONITOR_EXTENDED_ATTRIBUTES >= 1
// structure for getxattr() call
struct getxattr_args { 
        char path_l_[PADL_(user_addr_t)]; user_addr_t path; char path_r_[PADR_(user_addr_t)];
        char attrname_l_[PADL_(user_addr_t)]; user_addr_t attrname; char attrname_r_[PADR_(user_addr_t)];
        char value_l_[PADL_(user_addr_t)]; user_addr_t value; char value_r_[PADR_(user_addr_t)];
        char size_l_[PADL_(user_size_t)]; user_size_t size; char size_r_[PADR_(user_size_t)];
        char position_l_[PADL_(uint32_t)]; uint32_t position; char position_r_[PADR_(uint32_t)];
        char options_l_[PADL_(int)]; int options; char options_r_[PADR_(int)];
};
// structure for fgetxattr() call
struct fgetxattr_args {
        char fd_l_[PADL_(int)]; int fd; char fd_r_[PADR_(int)];
        char attrname_l_[PADL_(user_addr_t)]; user_addr_t attrname; char attrname_r_[PADR_(user_addr_t)];
        char value_l_[PADL_(user_addr_t)]; user_addr_t value; char value_r_[PADR_(user_addr_t)];
        char size_l_[PADL_(user_size_t)]; user_size_t size; char size_r_[PADR_(user_size_t)];
        char position_l_[PADL_(uint32_t)]; uint32_t position; char position_r_[PADR_(uint32_t)];
        char options_l_[PADL_(int)]; int options; char options_r_[PADR_(int)];
};
// structure for setxattr() call
struct setxattr_args {
        char path_l_[PADL_(user_addr_t)]; user_addr_t path; char path_r_[PADR_(user_addr_t)];
        char attrname_l_[PADL_(user_addr_t)]; user_addr_t attrname; char attrname_r_[PADR_(user_addr_t)];
        char value_l_[PADL_(user_addr_t)]; user_addr_t value; char value_r_[PADR_(user_addr_t)];
        char size_l_[PADL_(user_size_t)]; user_size_t size; char size_r_[PADR_(user_size_t)];
        char position_l_[PADL_(uint32_t)]; uint32_t position; char position_r_[PADR_(uint32_t)];
        char options_l_[PADL_(int)]; int options; char options_r_[PADR_(int)];
};
// structure for fsetxattr() call
struct fsetxattr_args {
        char fd_l_[PADL_(int)]; int fd; char fd_r_[PADR_(int)];
        char attrname_l_[PADL_(user_addr_t)]; user_addr_t attrname; char attrname_r_[PADR_(user_addr_t)];
        char value_l_[PADL_(user_addr_t)]; user_addr_t value; char value_r_[PADR_(user_addr_t)];
        char size_l_[PADL_(user_size_t)]; user_size_t size; char size_r_[PADR_(user_size_t)];
        char position_l_[PADL_(uint32_t)]; uint32_t position; char position_r_[PADR_(uint32_t)];
        char options_l_[PADL_(int)]; int options; char options_r_[PADR_(int)];
};
// structure for removexattr() call
struct removexattr_args {
        char path_l_[PADL_(user_addr_t)]; user_addr_t path; char path_r_[PADR_(user_addr_t)];
        char attrname_l_[PADL_(user_addr_t)]; user_addr_t attrname; char attrname_r_[PADR_(user_addr_t)];
        char options_l_[PADL_(int)]; int options; char options_r_[PADR_(int)];
};
// structure for fremovexattr() call
struct fremovexattr_args {
        char fd_l_[PADL_(int)]; int fd; char fd_r_[PADR_(int)];
        char attrname_l_[PADL_(user_addr_t)]; user_addr_t attrname; char attrname_r_[PADR_(user_addr_t)];
        char options_l_[PADL_(int)]; int options; char options_r_[PADR_(int)];
};
// structure for listxattr() call
struct listxattr_args {
        char path_l_[PADL_(user_addr_t)]; user_addr_t path; char path_r_[PADR_(user_addr_t)];
        char namebuf_l_[PADL_(user_addr_t)]; user_addr_t namebuf; char namebuf_r_[PADR_(user_addr_t)];
        char bufsize_l_[PADL_(user_size_t)]; user_size_t bufsize; char bufsize_r_[PADR_(user_size_t)];
        char options_l_[PADL_(int)]; int options; char options_r_[PADR_(int)];
};
// structure for flistxattr() call
struct flistxattr_args {
        char fd_l_[PADL_(int)]; int fd; char fd_r_[PADR_(int)];
        char namebuf_l_[PADL_(user_addr_t)]; user_addr_t namebuf; char namebuf_r_[PADR_(user_addr_t)];
        char bufsize_l_[PADL_(user_size_t)]; user_size_t bufsize; char bufsize_r_[PADR_(user_size_t)];
        char options_l_[PADL_(int)]; int options; char options_r_[PADR_(int)];
};
#endif

typedef int32_t	sy_call_t (struct proc *, void *, int *);
typedef void	sy_munge_t (const void *, void *);

// TIGER structure is different
// xnu/bsd/sys/sysent.h
#if BUILDING_ON_TIGER >= 1
struct sysent {          /* system call table */
        int16_t         sy_narg;        /* number of args */
        int8_t          sy_cancel;      /* funnel type */
        int8_t          sy_funnel;      /* funnel type */
        sy_call_t       *sy_call;       /* implementing function */
        sy_munge_t      *sy_arg_munge32; /* system call aguments munger for 32-bit process */
        sy_munge_t      *sy_arg_munge64; /* system call aguments munger for 64-bit process */
        int32_t         sy_return_type; /* system call return types */
};
#else
// this is for Leopard
// found in xnu-1228.7.58/bsd/sys/sysent.h
struct sysent {
	int16_t		sy_narg;		/* number of arguments */
	int8_t		reserved;		/* unused value */
	int8_t		sy_flags;		/* call flags */
	sy_call_t	*sy_call;		/* implementing function */
	sy_munge_t	*sy_arg_munge32;	/* munge system call arguments for 32-bit processes */
	sy_munge_t	*sy_arg_munge64;	/* munge system call arguments for 64-bit processes */
	int32_t		sy_return_type; /* return type */
	uint16_t	sy_arg_bytes;	/* The size of all arguments for 32-bit system calls, in bytes */
};
#endif


// ripped from xnu/bsd/sys/proc_internal.h
// Needed so we can access the proc structure passed to the syscall
// I had to comment some fields because other kernel includes would be needed
// This should be valid since we are not doing any copies of this structure (just having it's definition to avoid dereference pointer to incomplete types)
// else things might go wrong (big kabooommm)
#if BUILDING_ON_TIGER >=1
struct proc {
        LIST_ENTRY(proc) p_list;        /* List of all processes. */

        /* substructures: */
        struct  ucred *p_ucred;         /* Process owner's identity. */
        struct  filedesc *p_fd;         /* Ptr to open files structure. */
        struct  pstats *p_stats;        /* Accounting/statistics (PROC ONLY). */
        struct  plimit *p_limit;        /* Process limits. */
        struct  sigacts *p_sigacts;     /* Signal actions, state (PROC ONLY). */

#define p_rlimit        p_limit->pl_rlimit

        int     p_flag;                 /* P_* flags. */
        char    p_stat;                 /* S* process status. */
        char    p_shutdownstate;
        char    p_pad1[2];

        pid_t   p_pid;                  /* Process identifier. */ // <- this is want we really want :)
        LIST_ENTRY(proc) p_pglist;      /* List of processes in pgrp. */
        struct  proc *p_pptr;           /* Pointer to parent process. */
        LIST_ENTRY(proc) p_sibling;     /* List of sibling processes. */
        LIST_HEAD(, proc) p_children;   /* Pointer to list of children. */

/* The following fields are all zeroed upon creation in fork. */
#define p_startzero     p_oppid

        pid_t   p_oppid;         /* Save parent pid during ptrace. XXX */
        int     p_dupfd;         /* Sideways return value from fdopen. XXX */

        /* scheduling */
        u_int   p_estcpu;        /* Time averaged value of p_cpticks. */
        int     p_cpticks;       /* Ticks of cpu time. */
        fixpt_t p_pctcpu;        /* %cpu for this process during p_swtime */
        void    *p_wchan;        /* Sleep address. */
        char    *p_wmesg;        /* Reason for sleep. */
        u_int   p_swtime;        /* DEPRECATED (Time swapped in or out.) */
#define p_argslen p_swtime       /* Length of process arguments. */
        u_int   p_slptime;       /* Time since last blocked. */

        struct  itimerval p_realtimer;  /* Alarm timer. */
        struct  timeval p_rtime;        /* Real time. */
        u_quad_t p_uticks;              /* Statclock hits in user mode. */
        u_quad_t p_sticks;              /* Statclock hits in system mode. */
        u_quad_t p_iticks;              /* Statclock hits processing intr. */

        int     p_traceflag;            /* Kernel trace points. */
        struct  vnode *p_tracep;        /* Trace to vnode. */

        sigset_t p_siglist;             /* DEPRECATED. */
        struct  vnode *p_textvp;        /* Vnode of executable. */

/* End area that is zeroed on creation. */
#define p_endzero       p_hash.le_next

        /*
         * Not copied, not zero'ed.
         * Belongs after p_pid, but here to avoid shifting proc elements.
         */
        LIST_ENTRY(proc) p_hash;        /* Hash chain. */
        TAILQ_HEAD( ,eventqelt) p_evlist;

/* The following fields are all copied upon creation in fork. */
#define p_startcopy     p_sigmask

        sigset_t p_sigmask;             /* DEPRECATED */
        sigset_t p_sigignore;   /* Signals being ignored. */
        sigset_t p_sigcatch;    /* Signals being caught by user. */

        u_char  p_priority;     /* Process priority. */
        u_char  p_usrpri;       /* User-priority based on p_cpu and p_nice. */
        char    p_nice;         /* Process "nice" value. */
        char    p_comm[MAXCOMLEN+1];
        char    p_name[(2*MAXCOMLEN)+1];
        struct  pgrp *p_pgrp;   /* Pointer to process group. */

/* End area that is copied on creation. */
#define p_endcopy       p_xstat

        u_short p_xstat;        /* Exit status for wait; also stop signal. */
        u_short p_acflag;       /* Accounting flags. */
        struct  rusage *p_ru;   /* Exit information. XXX */

        int     p_debugger;     /* 1: can exec set-bit programs if suser */

        void    *task;                  /* corresponding task */
        void    *sigwait_thread;        /* 'thread' holding sigwait */
        char    signal_lock[72];
        boolean_t        sigwait;       /* indication to suspend */
        void    *exit_thread;           /* Which thread is exiting? */
        user_addr_t user_stack;         /* where user stack was allocated */
        void * exitarg;                 /* exit arg for proc terminate */
        void * vm_shm;                  /* for sysV shared memory */
        int  p_argc;                    /* saved argc for sysctl_procargs() */
        int             p_vforkcnt;             /* number of outstanding vforks */
    void *  p_vforkact;     /* activation running this vfork proc */
        TAILQ_HEAD( , uthread) p_uthlist; /* List of uthreads  */
        /* Following fields are info from SIGCHLD */
        pid_t   si_pid;
        u_short si_status;
        u_short si_code;
        uid_t   si_uid;
        TAILQ_HEAD( , aio_workq_entry ) aio_activeq; /* active async IO requests */
        int             aio_active_count;       /* entries on aio_activeq */
        TAILQ_HEAD( , aio_workq_entry ) aio_doneq;       /* completed async IO requests */
        int             aio_done_count;         /* entries on aio_doneq */

// COMMENTED OUT BECAUSE WE WOULD NEED OTHER KERNEL INCLUDES
//        struct klist p_klist;  /* knote list */
//        lck_mtx_t       p_mlock;        /* proc lock to protect evques */
//        lck_mtx_t       p_fdmlock;      /* proc lock to protect evques */
        unsigned int p_fdlock_pc[4];
        unsigned int p_fdunlock_pc[4];
        int             p_fpdrainwait;
        unsigned int            p_lflag;                /* local flags */
        unsigned int            p_ladvflag;             /* local adv flags*/
        unsigned int            p_internalref;  /* temp refcount field */
#if DIAGNOSTIC
#if SIGNAL_DEBUG
        unsigned int lockpc[8];
        unsigned int unlockpc[8];
#endif /* SIGNAL_DEBUG */
#endif /* DIAGNOSTIC */
};
#else
// FOR LEOPARD
struct  proc {
        LIST_ENTRY(proc) p_list;                /* List of all processes. */

        pid_t           p_pid;                  /* Process identifier. (static)*/
        void *          task;                   /* corresponding task (static)*/
        struct  proc *  p_pptr;                 /* Pointer to parent process.(LL) */
        pid_t           p_ppid;                 /* process's parent pid number */
        pid_t           p_pgrpid;               /* process group id of the process (LL)*/

//        lck_mtx_t       p_mlock;                /* mutex lock for proc */

        char            p_stat;                 /* S* process status. (PL)*/
        char            p_shutdownstate;
        char            p_kdebug;               /* P_KDEBUG eq (CC)*/
        char            p_btrace;               /* P_BTRACE eq (CC)*/

        LIST_ENTRY(proc) p_pglist;              /* List of processes in pgrp.(PGL) */
        LIST_ENTRY(proc) p_sibling;             /* List of sibling processes. (LL)*/
        LIST_HEAD(, proc) p_children;           /* Pointer to list of children. (LL)*/
        TAILQ_HEAD( , uthread) p_uthlist;       /* List of uthreads  (PL) */

        LIST_ENTRY(proc) p_hash;                /* Hash chain. (LL)*/
        TAILQ_HEAD( ,eventqelt) p_evlist;       /* (PL) */

//        lck_mtx_t       p_fdmlock;              /* proc lock to protect fdesc */

        /* substructures: */
        kauth_cred_t    p_ucred;                /* Process owner's identity. (PL) */
        struct  filedesc *p_fd;                 /* Ptr to open files structure. (PFDL) */
        struct  pstats *p_stats;                /* Accounting/statistics (PL). */
        struct  plimit *p_limit;                /* Process limits.(PL) */

        struct  sigacts *p_sigacts;             /* Signal actions, state (PL) */
//        lck_spin_t      p_slock;                /* spin lock for itimer/profil protection */

#define p_rlimit        p_limit->pl_rlimit

        struct  plimit *p_olimit;               /* old process limits  - not inherited by child  (PL) */
        unsigned int    p_flag;                 /* P_* flags. (atomic bit ops) */
        unsigned int    p_lflag;                /* local flags  (PL) */
        unsigned int    p_listflag;             /* list flags (LL) */
        unsigned int    p_ladvflag;             /* local adv flags (atomic) */
        int             p_refcount;             /* number of outstanding users(LL) */
        int             p_childrencnt;          /* children holding ref on parent (LL) */
        int             p_parentref;            /* children lookup ref on parent (LL) */

        pid_t           p_oppid;                /* Save parent pid during ptrace. XXX */
        u_int           p_xstat;                /* Exit status for wait; also stop signal. */

#ifdef _PROC_HAS_SCHEDINFO_
        /* may need cleanup, not used */
        u_int           p_estcpu;               /* Time averaged value of p_cpticks.(used by aio and proc_comapre) */
        fixpt_t         p_pctcpu;               /* %cpu for this process during p_swtime (used by aio)*/
        u_int           p_slptime;              /* used by proc_compare */
#endif /* _PROC_HAS_SCHEDINFO_ */

        struct  itimerval p_realtimer;          /* Alarm timer. (PSL) */
        struct  timeval p_rtime;                /* Real time.(PSL)  */
        struct  itimerval p_vtimer_user;        /* Virtual timers.(PSL)  */
        struct  itimerval p_vtimer_prof;        /* (PSL) */

        struct  timeval p_rlim_cpu;             /* Remaining rlim cpu value.(PSL) */
        int             p_debugger;             /*  NU 1: can exec set-bit programs if suser */
        boolean_t       sigwait;        /* indication to suspend (PL) */
        void    *sigwait_thread;        /* 'thread' holding sigwait(PL)  */
        void    *exit_thread;           /* Which thread is exiting(PL)  */
        int     p_vforkcnt;             /* number of outstanding vforks(PL)  */
        void *  p_vforkact;             /* activation running this vfork proc)(static)  */
        int     p_fpdrainwait;          /* (PFDL) */
        pid_t   p_contproc;     /* last PID to send us a SIGCONT (PL) */

        /* Following fields are info from SIGCHLD (PL) */
        pid_t   si_pid;                 /* (PL) */
        u_int   si_status;              /* (PL) */
        u_int   si_code;                /* (PL) */
        uid_t   si_uid;                 /* (PL) */

        void * vm_shm;                  /* (SYSV SHM Lock) for sysV shared memory */

#if CONFIG_DTRACE
        user_addr_t                     p_dtrace_argv;                  /* (write once, read only after that) */
        user_addr_t                     p_dtrace_envp;                  /* (write once, read only after that) */
        lck_mtx_t                       p_dtrace_sprlock;               /* sun proc lock emulation */
        int                             p_dtrace_probes;                /* (PL) are there probes for this proc? */
        u_int                           p_dtrace_count;                 /* (sprlock) number of DTrace tracepoints */
        struct dtrace_ptss_page*        p_dtrace_ptss_pages;            /* (sprlock) list of user ptss pages */
        struct dtrace_ptss_page_entry*  p_dtrace_ptss_free_list;        /* (atomic) list of individual ptss entries */
        struct dtrace_helpers*          p_dtrace_helpers;               /* (dtrace_lock) DTrace per-proc private */
        struct dof_ioctl_data*          p_dtrace_lazy_dofs;             /* (sprlock) unloaded dof_helper_t's */
#endif /* CONFIG_DTRACE */

/* XXXXXXXXXXXXX BCOPY'ed on fork XXXXXXXXXXXXXXXX */
/* The following fields are all copied upon creation in fork. */
#define p_startcopy     p_argslen

        u_int   p_argslen;       /* Length of process arguments. */
        int     p_argc;                 /* saved argc for sysctl_procargs() */
        user_addr_t user_stack;         /* where user stack was allocated */
        struct  vnode *p_textvp;        /* Vnode of executable. */
        off_t   p_textoff;              /* offset in executable vnode */

        sigset_t p_sigmask;             /* DEPRECATED */
        sigset_t p_sigignore;   /* Signals being ignored. (PL) */
        sigset_t p_sigcatch;    /* Signals being caught by user.(PL)  */

        u_char  p_priority;     /* (NU) Process priority. */
        u_char  p_resv0;        /* (NU) User-priority based on p_cpu and p_nice. */
        char    p_nice;         /* Process "nice" value.(PL) */
        u_char  p_resv1;        /* (NU) User-priority based on p_cpu and p_nice. */

#if CONFIG_MACF
        int     p_mac_enforce;                  /* MAC policy enforcement control */
#endif

        char    p_comm[MAXCOMLEN+1];
        char    p_name[(2*MAXCOMLEN)+1];        /* PL */

        struct  pgrp *p_pgrp;   /* Pointer to process group. (LL) */
        int             p_iopol_disk;   /* disk I/O policy (PL) */
        uint32_t        p_csflags;      /* flags for codesign (PL) */

/* End area that is copied on creation. */
/* XXXXXXXXXXXXX End of BCOPY'ed on fork (AIOLOCK)XXXXXXXXXXXXXXXX */
#define p_endcopy       aio_active_count
        int             aio_active_count;       /* entries on aio_activeq */
        int             aio_done_count;         /* entries on aio_doneq */
        TAILQ_HEAD( , aio_workq_entry ) aio_activeq; /* active async IO requests */
        TAILQ_HEAD( , aio_workq_entry ) aio_doneq;       /* completed async IO requests */

//        struct klist p_klist;  /* knote list (PL ?)*/

        struct  rusage *p_ru;   /* Exit information. (PL) */
        thread_t        p_signalholder;
        thread_t        p_transholder;

        /* DEPRECATE following field  */
        u_short p_acflag;       /* Accounting flags. */

        struct lctx *p_lctx;            /* Pointer to login context. */
        LIST_ENTRY(proc) p_lclist;      /* List of processes in lctx. */
        user_addr_t     p_threadstart;          /* pthread start fn */
        user_addr_t     p_wqthread;             /* pthread workqueue fn */
        int     p_pthsize;                      /* pthread size */
        void *  p_wqptr;                        /* workq ptr */
        int     p_wqsize;                       /* allocated size */
//        lck_mtx_t       p_wqlock;               /* lock to protect work queue */
        struct  timeval p_start;                /* starting time */
        void *  p_rcall;
        int             p_ractive;
#if DIAGNOSTIC
        unsigned int p_fdlock_pc[4];
        unsigned int p_fdunlock_pc[4];
#if SIGNAL_DEBUG
        unsigned int lockpc[8];
        unsigned int unlockpc[8];
#endif /* SIGNAL_DEBUG */
#endif /* DIAGNOSTIC */
};
#endif